import React from "react";
import './highlight.css'
const markTagOffset = 13;
export default class Highlight extends React.Component {
    constructor(props) {
        super(props);
        this.historyRanges = [];
        this.state={}
    }

    onMouseUpHandler = (index) => {
        const selectionObj = (window.getSelection && window.getSelection());
        if (selectionObj.isCollapsed) return;
        const { position, anchorNode, focusOffset, anchorOffset, focusNode, selection } = this.extractDetailsFrom(selectionObj);
        let forward = this.isForward(position, anchorNode, focusOffset, anchorOffset);
        let { parentNode, focusParentNode, focusFound, endOffset, anchorFound, startOffset } = this.setInitialValues(anchorNode, focusNode);
        ({ parentNode, focusParentNode } = this.substituteParentAndAnchorNode(anchorNode, parentNode, focusNode, focusParentNode));
        if (parentNode != focusParentNode) return;

        for (let i = 0; i < parentNode.childNodes.length; i++) {
            if (focusFound) break;
            let node = this.substituteNode(parentNode, i);
            ({ endOffset, focusFound } = this.updateEndOffset(node, focusNode, endOffset, focusOffset, focusFound));
        }

        for (let j = 0; j < parentNode.childNodes.length; j++) {
            if (anchorFound) break;
            let node = this.substituteNode(parentNode, i);
            ({ startOffset, anchorFound } = this.updateStartOffset(node, anchorNode, startOffset, anchorOffset, anchorFound));
        }
        this.callSelectionHandler(startOffset, endOffset, forward, selection, index);
    }

    selectionHandler = (selection, selectionStart, selectionEnd, paragraphIndex) => {
        let paragraph = this.state.paragraphs ? [...this.state.paragraphs] : [...this.props.paragraphs];
        let paragraphString = this.props.paragraphs[paragraphIndex];
        this.mergeRanges({start: selectionStart, end: selectionEnd}, paragraphIndex);
        paragraphString = this.updateParagraphString(paragraphIndex, paragraphString);
        paragraph[paragraphIndex] = paragraphString;
        this.setState({paragraphs: paragraph})
    }


    mergeRanges = (range, paraIndex) => {
        if (this.historyRanges[paraIndex] == null) this.historyRanges[paraIndex] = [];
        let mid = Math.floor((this.historyRanges[paraIndex].length) / 2);
        let index = this.historyRanges[paraIndex].length > 0 ? this.binarySearch(0, this.historyRanges[paraIndex].length - 1, mid, range, paraIndex) : mid;
        let merged = this.finallyMerge(paraIndex, merged, range);
        if (!merged) this.historyRanges[paraIndex] = [...this.historyRanges[paraIndex].slice(0, index), range, ...this.historyRanges[paraIndex].slice(index, this.historyRanges[paraIndex].length)];
    }

    mergeifMergeable = (source, target) => {
        let howToMerge = (source['start'] <= target['start'] && target['start'] <= source['end'] ? 1 :
            target['end'] <= source['end'] && target['end'] >= source['start'] ? 2 :
                target['start'] <= source['start'] && target['end'] >= source['end'] ? 3 : 0);
        if (howToMerge === 1) {
            return {start: source['start'], end: Math.max(target['end'], source['end'])}
        } else if (howToMerge === 2) {
            return {start: Math.min(source['start'], target['start']), end: source['end']}
        } else if (howToMerge === 3) {
            return target;
        }
        return false;
    }

    binarySearch = (start, end, mid, newEle, paraIndex) => {

        while (start < end) {
            let ele = this.historyRanges[paraIndex][mid];
            if (ele['start'] === newEle['start']) {
                return mid;
            } else if (ele['start'] > newEle['start']) {
                end = mid - 1
                if (end < start) break;
                mid = Math.ceil((start + end) / 2);
            } else {
                start = mid + 1;
                if (start > end) break;
                mid = Math.ceil((start + end) / 2);
            }
        }
        if (newEle['start'] > this.historyRanges[paraIndex][mid]['start']) {
            return mid + 1;
        } else {
            return mid;
        }
    }

    finallyMerge(paraIndex, range) {
        let merged = false;
        for (let i = 0; i < this.historyRanges[paraIndex].length; i++) {
            let merge = false;
            if (merged) {
                merge = this.mergeifMergeable(this.historyRanges[paraIndex][i], this.historyRanges[paraIndex][i - 1]);
            } else {
                merge = this.mergeifMergeable(this.historyRanges[paraIndex][i], range);
            }
            if (merge) {
                this.historyRanges[paraIndex][i] = merge;
                merged = true;
            }
        }
        return merged;
    }

    updateParagraphString(paragraphIndex, paragraphString) {
        this.historyRanges[paragraphIndex].forEach((range, index) => {
            let start = range['start'] + (index * markTagOffset);
            let end = range['end'] + (index * markTagOffset);
            paragraphString = `${paragraphString.slice(0, start)}<mark>${paragraphString.slice(start, end + 1)}</mark>${paragraphString.slice(end + 1, paragraphString.length)}`;
        });
        return paragraphString;
    }

    callSelectionHandler(startOffset, endOffset, forward, selection, index) {
        let selectionStart = startOffset;
        let selectionEnd = endOffset;
        ({ selectionStart, selectionEnd } = this.updateSelectionEndpoints(forward, selectionStart, endOffset, selectionEnd, startOffset));

        this.selectionHandler(selection, selectionStart, selectionEnd - 1, index);
    }

    updateSelectionEndpoints(forward, selectionStart, endOffset, selectionEnd, startOffset) {
        if (!forward) {
            selectionStart = endOffset;
            selectionEnd = startOffset;
        }
        return { selectionStart, selectionEnd };
    }

    updateStartOffset(node, anchorNode, startOffset, anchorOffset, anchorFound) {
        if (node != anchorNode) {
            startOffset += node.nodeType == 1 ? node.innerText.length : node.length;
        }
        if (node == anchorNode) {
            startOffset += anchorOffset;
            anchorFound = true;
        }
        return { startOffset, anchorFound };
    }

    updateEndOffset(node, focusNode, endOffset, focusOffset, focusFound) {
        if (node !== focusNode) {
            endOffset += node.nodeType === 1 ? node.innerText.length : node.length;
        }
        if (node === focusNode) {
            endOffset += focusOffset;
            focusFound = true;
        }
        return { endOffset, focusFound };
    }

    substituteNode(parentNode, i) {
        let node = parentNode.childNodes[i];
        if (node.nodeName === 'MARK') {
            node = node.childNodes[0];
        }
        return node;
    }

    substituteParentAndAnchorNode(anchorNode, parentNode, focusNode, focusParentNode) {
        if (anchorNode.parentElement.nodeName === "MARK") {
            parentNode = parentNode.parentNode;
        }
        if (focusNode.parentElement.nodeName === "MARK") {
            focusParentNode = focusParentNode.parentNode;
        }
        return { parentNode, focusParentNode };
    }

    setInitialValues(anchorNode, focusNode) {
        let endOffset = 0;
        let startOffset = 0;
        let focusFound = false;
        let anchorFound = false;
        let parentNode = anchorNode.parentNode;
        let focusParentNode = focusNode.parentNode;
        return { parentNode, focusParentNode, focusFound, endOffset, anchorFound, startOffset };
    }

    isForward(position, anchorNode, focusOffset, anchorOffset) {
        let forward = false;
        if (position === anchorNode.DOCUMENT_POSITION_FOLLOWING) {
            forward = true;
        } else if (position === 0) {
            forward = (focusOffset - anchorOffset) > 0;
        }
        return forward;
    }

    extractDetailsFrom(selectionObj) {
        const selection = selectionObj.toString();
        const anchorOffset = selectionObj.anchorOffset;
        const focusOffset = selectionObj.focusOffset;
        const anchorNode = selectionObj.anchorNode;
        const focusNode = selectionObj.focusNode;
        const position = anchorNode.compareDocumentPosition(focusNode);
        return { position, anchorNode, focusOffset, anchorOffset, focusNode, selection };
    }

    render() {
        let p = this.state.paragraphs || this.props.paragraphs;
        return(
            <>
                {
                    p.map((paragraph,i) => {
                        return <p dangerouslySetInnerHTML={{__html: paragraph}}
                                  key={i} className={`${this.props.className}`} id="" onMouseUp={(e) => {
                            e.preventDefault();
                            this.onMouseUpHandler(i)
                        }}/>
                    })
                }
            </>
        )
    }

}